 /*@@
   @file      Whisky_PPM.F90
   @date      Sun Feb 10 16:53:29 2002
   @author    Ian Hawke, Toni Font, Luca Baiotti, Frank Loeffler
   @desc 
   Routines to do PPM reconstruction.
   @enddesc 
 @@*/

#include "cctk.h"
#include "cctk_Arguments.h"
#include "cctk_Parameters.h"

 /*@@
   @routine    SimplePPM_1d
   @date       Thu Feb 14 19:08:52 2002
   @author     Ian Hawke, Toni Font
   @desc 
   The simple PPM reconstruction routine that applies along
   each one dimensional slice.
   @enddesc 
   @calls     
   @calledby   
   @history 
   Written in frustration when IH couldn''t get Toni''s original code 
   to work.
   @endhistory 

@@*/

subroutine SimplePPM_1d(handle,nx,dx,rho,velx,vely,velz,eps,rhominus,&
     velxminus,velyminus,velzminus,epsminus,rhoplus,velxplus,velyplus,&
     velzplus,epsplus)

  USE Whisky_Scalars

  implicit none

  DECLARE_CCTK_PARAMETERS

#ifdef _EOS_BASE_INC_
#undef _EOS_BASE_INC_
#endif
#include "EOS_Base.inc"

  CCTK_INT :: handle,nx
  CCTK_REAL :: dx
  CCTK_REAL, dimension(nx) :: rho,velx,vely,velz,eps
  CCTK_REAL, dimension(nx) :: rhominus,velxminus,velyminus,velzminus,epsminus
  CCTK_REAL, dimension(nx) :: rhoplus,velxplus,velyplus,velzplus,epsplus

  CCTK_INT :: i,s
  CCTK_REAL, dimension(nx) :: drho,dvelx,dvely,dvelz,deps
  CCTK_REAL, dimension(nx) :: dmrho,dmvelx,dmvely,dmvelz,dmeps
  CCTK_REAL, dimension(nx) :: press,dpress,d2rho,tilde_flatten
  CCTK_REAL :: dpress2,dvel,w,flatten,eta,etatilde

!!$  Loop to set the pressure

  do i = 1, nx
    press(i) = EOS_Pressure(handle, rho(i), eps(i))
  end do

!!$  Average slopes delta_m(a). See (1.7) of Colella and Woodward, p.178
!!$  This is the expression for an even grid.

  do i = 2, nx - 1
    drho(i) = 0.5d0 * (rho(i+1) - rho(i-1))
    dvelx(i) = 0.5d0 * (velx(i+1) - velx(i-1))
    dvely(i) = 0.5d0 * (vely(i+1) - vely(i-1))
    dvelz(i) = 0.5d0 * (velz(i+1) - velz(i-1))
    deps(i) = 0.5d0 * (eps(i+1) - eps(i-1))
    dpress(i) = press(i+1) - press(i-1)
    d2rho(i) = (rho(i+1) - 2.d0 * rho(i) + rho(i-1))! / 6.d0 / dx / dx
    ! since we use d2rho only for the condition d2rho(i+1)*d2rhoi(i-1)<0 
    ! the denominator is not necessary
  end do

!!$  Steepened slope. See (1.8) of Colella and Woodward, p.178

  do i = 2, nx - 1
#define STEEP(x,dx,dmx)                                              \
    if ( (x(i+1) - x(i)) * (x(i) - x(i-1)) > 0.d0 ) then           &&\
      dmx(i) = sign(1.d0, dx(i)) *                                   \
           min(abs(dx(i)), 2.d0 * abs(x(i) - x(i-1)),                \
           2.d0 * abs(x(i+1) - x(i)))                              &&\
    else                                                           &&\
      dmx(i) = 0.d0                                                &&\
    end if
    STEEP(rho, drho, dmrho)
    STEEP(velx, dvelx, dmvelx)
    STEEP(vely, dvely, dmvely)
    STEEP(velz, dvelz, dmvelz)
    STEEP(eps, deps, dmeps)
    end do

!!$  Initial boundary states. See (1.9) of Colella and Woodward, p.178

  do i = 2, nx-2
    rhoplus(i) = 0.5d0 * (rho(i) + rho(i+1)) + &
         (dmrho(i) - dmrho(i+1)) / 6.d0
    rhominus(i+1) = rhoplus(i)
    velxplus(i) = 0.5d0 * (velx(i) + velx(i+1)) + &
         (dmvelx(i) - dmvelx(i+1)) / 6.d0
    velxminus(i+1) = velxplus(i)
    velyplus(i) = 0.5d0 * (vely(i) + vely(i+1)) + &
         (dmvely(i) - dmvely(i+1)) / 6.d0
    velyminus(i+1) = velyplus(i)
    velzplus(i) = 0.5d0 * (velz(i) + velz(i+1)) + &
         (dmvelz(i) - dmvelz(i+1)) / 6.d0
    velzminus(i+1) = velzplus(i)
    epsplus(i) = 0.5d0 * (eps(i) + eps(i+1)) + &
         (dmeps(i) - dmeps(i+1)) / 6.d0
    epsminus(i+1) = epsplus(i)
  end do

!!$Discontinuity steepening. See (1.14-17) of C&W.
!!$This is the detect routine which mat be activated with the ppm_detect parameter
!!$Note that this part really also depends on the grid being even. 
!!$Note also that we don''t have access to the gas constant gamma.
!!$So this is just dropped from eq. (3.2) of C&W.
!!$We can get around this by just rescaling the constant k0 (ppm_k0 here).

  if (ppm_detect .ne. 0) then

    do i = 3, nx - 2
      if ((d2rho(i+1)*d2rho(i-1)<0.d0).and.(abs(rho(i+1)-rho(i-1)) - &
           ppm_epsilon_shock * min(abs(rho(i+1)), abs(rho(i-1))) > 0.d0)) then
        etatilde = (rho(i-2) - rho(i+2) + 4.d0 * drho(i)) / (drho(i) * 12.d0)
      else
        etatilde = 0.d0
      end if
      eta = max(0.d0, min(1.d0, ppm_eta1 * (etatilde - ppm_eta2)))
      if (ppm_k0 * abs(drho(i)) * min(press(i-1),press(i+1)) < &
           abs(dpress(i)) * min(rho(i-1), rho(i+1))) then
        eta = 0.d0
      end if
      rhominus(i) = rhominus(i) * (1.d0 - eta) + &
           (rho(i-1) + 0.5d0 * dmrho(i-1)) * eta
      rhoplus(i) = rhoplus(i) * (1.d0 - eta) + &
           (rho(i+1) - 0.5d0 * dmrho(i+1)) * eta
    end do

  end if

!!$  Zone flattening. See appendix of C&W, p. 197-8.

  do i = 3, nx - 2
    dpress2 = press(i+2) - press(i-2)
    dvel = velx(i+1) - velx(i-1)
    if ( (abs(dpress(i)) >  ppm_epsilon * min(press(i-1),press(i+1))) .and. &
         (dvel < 0.d0) ) then
      w = 1.d0
    else
      w = 0.d0
    end if
    if (abs(dpress2) < ppm_small) then
      tilde_flatten(i) = 1.d0
    else
      tilde_flatten(i) = max(0.d0, 1.d0 - w * max(0.d0, ppm_omega2 * &
           (dpress(i) / dpress2 - ppm_omega1)))
    end if
  end do


  if (PPM3) then
    do i = 3, nx - 2
      flatten = tilde_flatten(i)
      rhoplus(i) = flatten * rhoplus(i) + (1.d0 - flatten) * rho(i)
      rhominus(i) = flatten * rhominus(i) + (1.d0 - flatten) * rho(i)
      velxplus(i) = flatten * velxplus(i) + (1.d0 - flatten) * velx(i)
      velxminus(i) = flatten * velxminus(i) + (1.d0 - flatten) * velx(i)
      velyplus(i) = flatten * velyplus(i) + (1.d0 - flatten) * vely(i)
      velyminus(i) = flatten * velyminus(i) + (1.d0 - flatten) * vely(i)
      velzplus(i) = flatten * velzplus(i) + (1.d0 - flatten) * velz(i)
      velzminus(i) = flatten * velzminus(i) + (1.d0 - flatten) * velz(i)
      epsplus(i) = flatten * epsplus(i) + (1.d0 - flatten) * eps(i)
      epsminus(i) = flatten * epsminus(i) + (1.d0 - flatten) * eps(i)
    end do
  else  !!$ Really implement C&W, page 197; which requires stencil 4.
    do i = 4, nx - 3
      s=sign(1.d0, -dpress(i))
      flatten = max(tilde_flatten(i), tilde_flatten(i+s))  
      rhoplus(i) = flatten * rhoplus(i) + (1.d0 - flatten) * rho(i)
      rhominus(i) = flatten * rhominus(i) + (1.d0 - flatten) * rho(i)
      velxplus(i) = flatten * velxplus(i) + (1.d0 - flatten) * velx(i)
      velxminus(i) = flatten * velxminus(i) + (1.d0 - flatten) * velx(i)
      velyplus(i) = flatten * velyplus(i) + (1.d0 - flatten) * vely(i)
      velyminus(i) = flatten * velyminus(i) + (1.d0 - flatten) * vely(i)
      velzplus(i) = flatten * velzplus(i) + (1.d0 - flatten) * velz(i)
      velzminus(i) = flatten * velzminus(i) + (1.d0 - flatten) * velz(i)
      epsplus(i) = flatten * epsplus(i) + (1.d0 - flatten) * eps(i)
      epsminus(i) = flatten * epsminus(i) + (1.d0 - flatten) * eps(i)
    end do
  end if


!!$Monotonicity. See (1.10) of C&W.

  do i = whisky_stencil, nx - whisky_stencil + 1
    if ( (rhoplus(i) - rho(i)) * (rho(i) - rhominus(i)) .le. 0.d0 ) then
      rhominus(i) = rho(i)
      rhoplus(i) = rho(i)
    else if ( (rhoplus(i) - rhominus(i)) * (rho(i) - 0.5d0 * &
         (rhoplus(i) + rhominus(i))) > &
         (rhoplus(i) - rhominus(i))**2 / 6.d0 ) then
      rhominus(i) = 3.d0 * rho(i) - 2.d0 * rhoplus(i)
    else if ( (rhoplus(i) - rhominus(i)) * (rho(i) - 0.5d0 * &
         (rhoplus(i) + rhominus(i))) < &
         - (rhoplus(i) - rhominus(i))**2 / 6.d0 ) then
      rhoplus(i) = 3.d0 * rho(i) - 2.d0 * rhominus(i)
    end if
    if ( (velxplus(i) - velx(i)) * (velx(i) - velxminus(i)) .le. 0.d0 ) then
      velxminus(i) = velx(i)
      velxplus(i) = velx(i)
    else if ( (velxplus(i) - velxminus(i)) * (velx(i) - 0.5d0 * &
         (velxplus(i) + velxminus(i))) > &
         (velxplus(i) - velxminus(i))**2 / 6.d0 ) then
      velxminus(i) = 3.d0 * velx(i) - 2.d0 * velxplus(i)
    else if ( (velxplus(i) - velxminus(i)) * (velx(i) - 0.5d0 * &
         (velxplus(i) + velxminus(i))) < &
         - (velxplus(i) - velxminus(i))**2 / 6.d0 ) then
      velxplus(i) = 3.d0 * velx(i) - 2.d0 * velxminus(i)
    end if
    if ( (velyplus(i) - vely(i)) * (vely(i) - velyminus(i)) .le. 0.d0 ) then
      velyminus(i) = vely(i)
      velyplus(i) = vely(i)
    else if ( (velyplus(i) - velyminus(i)) * (vely(i) - 0.5d0 * &
         (velyplus(i) + velyminus(i))) > &
         (velyplus(i) - velyminus(i))**2 / 6.d0 ) then
      velyminus(i) = 3.d0 * vely(i) - 2.d0 * velyplus(i)
    else if ( (velyplus(i) - velyminus(i)) * (vely(i) - 0.5d0 * &
         (velyplus(i) + velyminus(i))) < &
         - (velyplus(i) - velyminus(i))**2 / 6.d0 ) then
      velyplus(i) = 3.d0 * vely(i) - 2.d0 * velyminus(i)
    end if
    if ( (velzplus(i) - velz(i)) * (velz(i) - velzminus(i)) .le. 0.d0 ) then
      velzminus(i) = velz(i)
      velzplus(i) = velz(i)
    else if ( (velzplus(i) - velzminus(i)) * (velz(i) - 0.5d0 * &
         (velzplus(i) + velzminus(i))) > &
         (velzplus(i) - velzminus(i))**2 / 6.d0 ) then
      velzminus(i) = 3.d0 * velz(i) - 2.d0 * velzplus(i)
    else if ( (velzplus(i) - velzminus(i)) * (velz(i) - 0.5d0 * &
         (velzplus(i) + velzminus(i))) < &
         - (velzplus(i) - velzminus(i))**2 / 6.d0 ) then
      velzplus(i) = 3.d0 * velz(i) - 2.d0 * velzminus(i)
    end if
    if ( (epsplus(i) - eps(i)) * (eps(i) - epsminus(i)) .le. 0.d0 ) then
      epsminus(i) = eps(i)
      epsplus(i) = eps(i)
    else if ( (epsplus(i) - epsminus(i)) * (eps(i) - 0.5d0 * &
         (epsplus(i) + epsminus(i))) > &
         (epsplus(i) - epsminus(i))**2 / 6.d0 ) then
      epsminus(i) = 3.d0 * eps(i) - 2.d0 * epsplus(i)
    else if ( (epsplus(i) - epsminus(i)) * (eps(i) - 0.5d0 * &
         (epsplus(i) + epsminus(i))) < &
         - (epsplus(i) - epsminus(i))**2 / 6.d0 ) then
      epsplus(i) = 3.d0 * eps(i) - 2.d0 * epsminus(i)
    end if
  end do

  return

end subroutine SimplePPM_1d

 /*@@
   @routine    SimplePPM_1dPolytype
   @date       Tue Mar 19 22:50:57 2002
   @author     Ian Hawke
   @desc 
   Currently the dummy version of above specialized to polytropes.
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

@@*/


subroutine SimplePPM_1dPolytype(handle,nx,dx,rho,velx,vely,velz,eps,rhominus,&
     velxminus,velyminus,velzminus,epsminus,rhoplus,velxplus,velyplus,&
     velzplus,epsplus)

  USE Whisky_Scalars

  implicit none

  DECLARE_CCTK_PARAMETERS

#ifdef _EOS_BASE_INC_
#undef _EOS_BASE_INC_
#endif
#include "EOS_Base.inc"

  CCTK_INT :: handle,nx
  CCTK_REAL :: dx
  CCTK_REAL, dimension(nx) :: rho,velx,vely,velz,eps
  CCTK_REAL, dimension(nx) :: rhominus,velxminus,velyminus,velzminus,epsminus
  CCTK_REAL, dimension(nx) :: rhoplus,velxplus,velyplus,velzplus,epsplus

  CCTK_INT :: i,s
  CCTK_REAL, dimension(nx) :: drho,dvelx,dvely,dvelz
  CCTK_REAL, dimension(nx) :: dmrho,dmvelx,dmvely,dmvelz
  CCTK_REAL, dimension(nx) :: press,dpress,d2rho,tilde_flatten
  CCTK_REAL :: dpress2,dvel,w,flatten,eta,etatilde


!!$  Loop to set the pressure and eps

  do i = 1, nx
    press(i) = EOS_Pressure(handle, rho(i), eps(i))
    eps(i) = EOS_SpecificIntEnergy(handle, rho(i), press(i))
  end do

!!$  Average slopes delta_m(a). See (1.7) of Colella and Woodward, p.178
!!$  This is the expression for an even grid.

  do i = 2, nx - 1
    drho(i) = 0.5d0 * (rho(i+1) - rho(i-1))
    dvelx(i) = 0.5d0 * (velx(i+1) - velx(i-1))
    dvely(i) = 0.5d0 * (vely(i+1) - vely(i-1))
    dvelz(i) = 0.5d0 * (velz(i+1) - velz(i-1))
    dpress(i) = press(i+1) - press(i-1)
    d2rho(i) = (rho(i+1) - 2.d0 * rho(i) + rho(i-1)) / 6.d0 / dx / dx
  end do

!!$  Steepened slope. See (1.8) of Colella and Woodward, p.178

  do i = 2, nx - 1
    if ( (rho(i+1) - rho(i)) * (rho(i) - rho(i-1)) > 0.d0 ) then
      dmrho(i) = sign(1.d0, drho(i)) * &
           min(abs(drho(i)), 2.d0 * abs(rho(i) - rho(i-1)), &
           2.d0 * abs(rho(i+1) - rho(i)))
    else
      dmrho(i) = 0.d0
    end if
    if ( (velx(i+1) - velx(i)) * (velx(i) - velx(i-1)) > 0.d0 ) then
      dmvelx(i) = sign(1.d0, dvelx(i)) * &
           min(abs(dvelx(i)), 2.d0 * abs(velx(i) - velx(i-1)), &
           2.d0 * abs(velx(i+1) - velx(i)))
    else
      dmvelx(i) = 0.d0
    end if
    if ( (vely(i+1) - vely(i)) * (vely(i) - vely(i-1)) > 0.d0 ) then
      dmvely(i) = sign(1.d0, dvely(i)) * &
           min(abs(dvely(i)), 2.d0 * abs(vely(i) - vely(i-1)), &
           2.d0 * abs(vely(i+1) - vely(i)))
    else
      dmvely(i) = 0.d0
    end if
    if ( (velz(i+1) - velz(i)) * (velz(i) - velz(i-1)) > 0.d0 ) then
      dmvelz(i) = sign(1.d0, dvelz(i)) * &
           min(abs(dvelz(i)), 2.d0 * abs(velz(i) - velz(i-1)), &
           2.d0 * abs(velz(i+1) - velz(i)))
    else
      dmvelz(i) = 0.d0
    end if
  end do

!!$  Initial boundary states. See (1.9) of Colella and Woodward, p.178

  do i = 2, nx-2
    rhoplus(i) = 0.5d0 * (rho(i) + rho(i+1)) + &
         (dmrho(i) - dmrho(i+1)) / 6.d0
    rhominus(i+1) = rhoplus(i)
    velxplus(i) = 0.5d0 * (velx(i) + velx(i+1)) + &
         (dmvelx(i) - dmvelx(i+1)) / 6.d0
    velxminus(i+1) = velxplus(i)
    velyplus(i) = 0.5d0 * (vely(i) + vely(i+1)) + &
         (dmvely(i) - dmvely(i+1)) / 6.d0
    velyminus(i+1) = velyplus(i)
    velzplus(i) = 0.5d0 * (velz(i) + velz(i+1)) + &
         (dmvelz(i) - dmvelz(i+1)) / 6.d0
    velzminus(i+1) = velzplus(i)
  end do

!!$Discontinuity steepening. See (1.14-17) of C&W.
!!$This is the detect routine which currently isn''t used
!!$Note that this part really, really, depends on the grid being even. 
!!$Note also that we don''t have access to the gas constant gamma.
!!$So this is just dropped from eq. (3.2) of C&W.
!!$We can get around this by just rescaling the constant k0 (ppm_k0 here).

  if (ppm_detect .ne. 0) then

    do i = 3, nx - 2
      if (-d2rho(i+1)*d2rho(i-1)*(rho(i+1)-rho(i-1)) - &
           ppm_epsilon_shock * min(abs(rho(i+1)), abs(rho(i-1))) > 0.d0) then
        etatilde = 0.5d0 * (rho(i+2) - rho(i-2) + 4.d0 * drho(i)) / drho(i)
      else
        etatilde = 0.d0
      end if
      eta = max(0.d0, min(1.d0, ppm_eta1 * (etatilde - ppm_eta2)))
      if (ppm_k0 * abs(drho(i)) * min(press(i-1),press(i+1)) < &
           abs(dpress(i)) * min(rho(i-1), rho(i+1))) then
        eta = 0.d0
      end if
      rhominus(i) = rhominus(i) * (1.d0 - eta) + &
           (rho(i-1) + 0.5d0 * dmrho(i-1)) * eta
      rhoplus(i) = rhoplus(i) * (1.d0 - eta) + &
           (rho(i+1) - 0.5d0 * dmrho(i+1)) * eta
    end do

  end if

!!$  Zone flattening. See appendix of C&W, p. 197-8.

  do i = 3, nx - 2
    dpress2 = press(i+2) - press(i-2)
    dvel = velx(i+1) - velx(i-1)
    if ( (abs(dpress(i)) >  ppm_epsilon * min(press(i-1),press(i+1))) .and. &
         (dvel < 0.d0) ) then
      w = 1.d0
    else
      w = 0.d0
    end if
    if (abs(dpress2) < ppm_small) then
      tilde_flatten(i) = 1.d0
    else
      tilde_flatten(i) = max(0.d0, 1.d0 - w * max(0.d0, ppm_omega2 * &
           (dpress(i) / dpress2 - ppm_omega1)))
    end if
  end do


  if (PPM3) then
    do i = 3, nx - 2
      flatten = tilde_flatten(i)
      rhoplus(i) = flatten * rhoplus(i) + (1.d0 - flatten) * rho(i)
      rhominus(i) = flatten * rhominus(i) + (1.d0 - flatten) * rho(i)
      velxplus(i) = flatten * velxplus(i) + (1.d0 - flatten) * velx(i)
      velxminus(i) = flatten * velxminus(i) + (1.d0 - flatten) * velx(i)
      velyplus(i) = flatten * velyplus(i) + (1.d0 - flatten) * vely(i)
      velyminus(i) = flatten * velyminus(i) + (1.d0 - flatten) * vely(i)
      velzplus(i) = flatten * velzplus(i) + (1.d0 - flatten) * velz(i)
      velzminus(i) = flatten * velzminus(i) + (1.d0 - flatten) * velz(i)
    end do
  else  !!$ Really implement C&W, page 197; which requires stencil 4.
    do i = 4, nx - 3
      s=sign(1.d0, -dpress(i))
      flatten = max(tilde_flatten(i), tilde_flatten(i+s))  
      rhoplus(i) = flatten * rhoplus(i) + (1.d0 - flatten) * rho(i)
      rhominus(i) = flatten * rhominus(i) + (1.d0 - flatten) * rho(i)
      velxplus(i) = flatten * velxplus(i) + (1.d0 - flatten) * velx(i)
      velxminus(i) = flatten * velxminus(i) + (1.d0 - flatten) * velx(i)
      velyplus(i) = flatten * velyplus(i) + (1.d0 - flatten) * vely(i)
      velyminus(i) = flatten * velyminus(i) + (1.d0 - flatten) * vely(i)
      velzplus(i) = flatten * velzplus(i) + (1.d0 - flatten) * velz(i)
      velzminus(i) = flatten * velzminus(i) + (1.d0 - flatten) * velz(i)
    end do
  end if

!!$Monotonicity. See (1.10) of C&W.

!!$ Knarf: changed loop from beeing from "2 to nx-2" to "3 to nx-2"
!!$        because vel?minus(2) is uninitialised here
  do i = 3, nx - 2
    if ( (rhoplus(i) - rho(i)) * (rho(i) - rhominus(i)) .le. 0.d0 ) then
      rhominus(i) = rho(i)
      rhoplus(i) = rho(i)
    else if ( (rhoplus(i) - rhominus(i)) * (rho(i) - 0.5d0 * &
         (rhoplus(i) + rhominus(i))) > &
         (rhoplus(i) - rhominus(i))**2 / 6.d0 ) then
      rhominus(i) = 3.d0 * rho(i) - 2.d0 * rhoplus(i)
    else if ( (rhoplus(i) - rhominus(i)) * (rho(i) - 0.5d0 * &
         (rhoplus(i) + rhominus(i))) < &
         - (rhoplus(i) - rhominus(i))**2 / 6.d0 ) then
      rhoplus(i) = 3.d0 * rho(i) - 2.d0 * rhominus(i)
    end if
    if ( (velxplus(i) - velx(i)) * (velx(i) - velxminus(i)) .le. 0.d0 ) then
      velxminus(i) = velx(i)
      velxplus(i) = velx(i)
    else if ( (velxplus(i) - velxminus(i)) * (velx(i) - 0.5d0 * &
         (velxplus(i) + velxminus(i))) > &
         (velxplus(i) - velxminus(i))**2 / 6.d0 ) then
      velxminus(i) = 3.d0 * velx(i) - 2.d0 * velxplus(i)
    else if ( (velxplus(i) - velxminus(i)) * (velx(i) - 0.5d0 * &
         (velxplus(i) + velxminus(i))) < &
         - (velxplus(i) - velxminus(i))**2 / 6.d0 ) then
      velxplus(i) = 3.d0 * velx(i) - 2.d0 * velxminus(i)
    end if
    if ( (velyplus(i) - vely(i)) * (vely(i) - velyminus(i)) .le. 0.d0 ) then
      velyminus(i) = vely(i)
      velyplus(i) = vely(i)
    else if ( (velyplus(i) - velyminus(i)) * (vely(i) - 0.5d0 * &
         (velyplus(i) + velyminus(i))) > &
         (velyplus(i) - velyminus(i))**2 / 6.d0 ) then
      velyminus(i) = 3.d0 * vely(i) - 2.d0 * velyplus(i)
    else if ( (velyplus(i) - velyminus(i)) * (vely(i) - 0.5d0 * &
         (velyplus(i) + velyminus(i))) < &
         - (velyplus(i) - velyminus(i))**2 / 6.d0 ) then
      velyplus(i) = 3.d0 * vely(i) - 2.d0 * velyminus(i)
    end if
    if ( (velzplus(i) - velz(i)) * (velz(i) - velzminus(i)) .le. 0.d0 ) then
      velzminus(i) = velz(i)
      velzplus(i) = velz(i)
    else if ( (velzplus(i) - velzminus(i)) * (velz(i) - 0.5d0 * &
         (velzplus(i) + velzminus(i))) > &
         (velzplus(i) - velzminus(i))**2 / 6.d0 ) then
      velzminus(i) = 3.d0 * velz(i) - 2.d0 * velzplus(i)
    else if ( (velzplus(i) - velzminus(i)) * (velz(i) - 0.5d0 * &
         (velzplus(i) + velzminus(i))) < &
         - (velzplus(i) - velzminus(i))**2 / 6.d0 ) then
      velzplus(i) = 3.d0 * velz(i) - 2.d0 * velzminus(i)
    end if

  end do

  return

end subroutine SimplePPM_1dPolytype
